@@ -28,8 +28,8 @@
 #include <assert.h>
 #include <stdbool.h>
 
-#define TCP_RECV_BUF_SIZE (16 * 1024)
-#define TCP_SEND_BUF_SIZE (4 * 1024)
+#define TCP_RECV_BUF_SIZE (65535 + 2)
+#define TCP_SEND_BUF_SIZE (65535 + 2)
 
 struct perf_sockaddr {
     union {
@@ -184,4 +184,6 @@ void perf_net_doh_stats_init();
 void perf_net_doh_stats_compile(struct perf_net_socket*);
 void perf_net_doh_stats_print();
 
+extern const char* perf_net_tls_sni;
+
 #endif
