@@ -61,7 +61,7 @@
 #define DEFAULT_MAX_OUTSTANDING (64 * 1024)
 #define DEFAULT_MAX_FALL_BEHIND 1000
 
-#define MAX_INPUT_DATA (64 * 1024)
+#define MAX_INPUT_DATA (64 * 1024) + 2
 
 #define TIMEOUT_CHECK_TIME 5000000
 
